// Copyright (c) 2022 Marceline Crmaer
// SPDX-License-Identifier: AGPL-3.0-or-later

use super::widgets::prelude::*;
use api::*;
use canary_script::*;
use canary_textwrap::{Content, Layout, TextCache};

use dialog::{DialogBodyStyle, DialogHeaderStyle};
use shell::Offset;
use text::{Label, LabelText};

pub struct NotificationStyle {
    pub header: DialogHeaderStyle,
    pub body: DialogBodyStyle,
    pub rounding: f32,
}

impl Default for NotificationStyle {
    fn default() -> Self {
        Self {
            header: DialogHeaderStyle {
                height: 12.0,
                ..Default::default()
            },
            body: Default::default(),
            rounding: THEME.metrics.surface_rounding,
        }
    }
}

pub struct NotificationPanel {
    panel: Panel,
    style: NotificationStyle,
    summary: Label,
    text_cache: TextCache,
    body: Content,
    body_layout: Layout,
    header_rect: Rect,
    body_rect: Rect,
}

impl PanelImpl for NotificationPanel {
    fn update(&mut self, dt: f32) {}

    fn draw(&mut self) {
        let ctx = DrawContext::new(self.panel);

        ctx.draw_partially_rounded_rect(
            CornerFlags::TOP,
            self.header_rect,
            self.style.rounding,
            self.style.header.color,
        );

        ctx.draw_partially_rounded_rect(
            CornerFlags::BOTTOM,
            self.body_rect,
            self.style.rounding,
            self.style.body.color,
        );

        self.summary.draw(&ctx);

        let ctx = ctx.with_offset(Vec2::new(5.0, 20.0));
        self.body_layout
            .draw(&self.text_cache, &ctx, 5.0, 8.0, THEME.palette.text);
    }

    fn on_resize(&mut self, new_size: Vec2) {
        let style = &self.style;
        let width = new_size.x;
        let body_height = new_size.y - style.header.height;
        let body_height = body_height.max(0.0);
        let header_size = Vec2::new(width, style.header.height);
        let body_size = Vec2::new(width, body_height);

        self.header_rect = Rect::from_xy_size(Vec2::ZERO, header_size);
        self.body_rect = Rect::from_xy_size(self.header_rect.bl(), body_size);

        let width = (new_size.x - 10.0) / 5.0;
        self.body_layout = self.body.layout(&mut self.text_cache, width);
    }

    fn on_cursor_event(&mut self, kind: CursorEventKind, at: Vec2) {}

    fn on_message(&mut self, msg: Message) {}
}

impl NotificationPanel {
    pub fn bind(panel: Panel, msg: Message) -> Box<dyn PanelImpl> {
        let msg = msg.to_vec();
        let msg: canary_notifications::Contents = serde_json::from_slice(&msg).unwrap();

        let style = NotificationStyle::default();
        let font = style.header.text_font;
        let text = msg.summary;
        let text = LabelText { font, text };
        let scale = style.header.height * style.header.text_scale_factor;
        let summary = Label::new(
            text,
            text::HorizontalAlignment::Left,
            scale,
            style.header.text_color,
            5.0,
            5.0,
            style.header.height * (1.0 - style.header.text_baseline),
        );

        let font = Font::new(crate::CONTENT_FONT);
        let text = msg.body.unwrap_or(String::new());
        let mut text_cache = TextCache::default();
        let body = Content::from_plain(&mut text_cache, font, &text);
        let body_layout = body.layout(&text_cache, 0.0);

        let header_rect = Default::default();
        let body_rect = Default::default();

        Box::new(Self {
            style,
            panel,
            summary,
            text_cache,
            body,
            body_layout,
            header_rect,
            body_rect,
        })
    }
}
